package wax-lib

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Theo.VarSource

Global unique variables.

Sourcetype 'kind t

The type of variables. The type parameter indicates the kind of value (bool, string, etc.).

Sourceval fresh : unit -> 'kind t

Create a fresh variable with a globally unique identifier.

Sourceval equal : 'kind t -> 'kind t -> bool
Sourceval compare : 'kind t -> 'kind t -> int
Sourceval hash : 'kind t -> int